#!/usr/bin/Rscript

# This file is part of Parco2015-nmelot.
#
# Parco2015-nmelot is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Parco2015-nmelot is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Parco2015-nmelot. If not, see <http://www.gnu.org/licenses/>.
#


options(error = function() {traceback(2); dump.frames; stopifnot(FALSE)})
## Load packages
library(ggplot2)
library(plyr)

## Read the frame and source the labelling script
data.frame = read.csv("table.csv")
source("labels.r")

## Show values within arithmetic expressions
show = function(x)
{
	print(x)
	return(x)
}

#gradient = function(colors = c("#FF55FF", "#FF4444", "#FFFF33", "#22FF22", "#11FFFF", "#0000FF"), grayscale = c("#FDFDFD", "#020202"), number = 10)
#gradient = function(colors = c("#0000FF", "#11FFFF", "#22FF22", "#FFFF33", "#FF4444", "#FF55FF"), grayscale = c("#FDFDFD", "#020202"), number = 10)
gradient = function(colors = c("#0000FF", "#11FFFF", "#22FF22", "#FFFF33", "#FF4444", "#00FF00"), grayscale = c("#888888", "#CCCCCC"), number = 10)
{
	## Steps is a vector of rgb colors the generated gradient must use
	## number is the number of colors to be generated

	segment = function(x, y, i, begin)
	{
		if(begin)
		{
			return(floor((i - 1) / ifelse(x > 1 , x - 1, 1) * (y - 1) ) + 1)
		}
		else
		{
			return(ceiling((i - 1) / ifelse(x > 1 , x - 1, 1) * (y - 1) ) + 1)
		}
	}

	ratio = function(x, y, i)
	{
		return((i - 1) / ifelse(x > 1 , x - 1, 1) * (y - 1) + 1 - segment(x, y, i, TRUE))
	}


	x = number
	y = length(colors)

	yy = length(grayscale)

	out = list()


	for(i in 1:number)
	{
		r = ratio(x, y, i)
		begin = segment(x, y, i, TRUE)
		end = segment(x, y, i, FALSE)
		#cat(paste(sep = "", "i: ", as.character(i), "; x: ", as.character(x), "; y = ", as.character(y), "; begin: ", as.character(begin), "; end: ", as.character(end), "; r: ", as.character(r)))

		rbegin = col2rgb(colors[begin], alpha = TRUE)["red",]
		gbegin = col2rgb(colors[begin], alpha = TRUE)["green",]
		bbegin = col2rgb(colors[begin], alpha = TRUE)["blue",]
		abegin = col2rgb(colors[begin], alpha = TRUE)["alpha",]

		rend = col2rgb(colors[end], alpha = TRUE)["red",]
		gend = col2rgb(colors[end], alpha = TRUE)["green",]
		bend = col2rgb(colors[end], alpha = TRUE)["blue",]
		aend = col2rgb(colors[end], alpha = TRUE)["alpha",]

		#cat(paste(sep="", "rbegin: ", as.character(rbegin), "; gbegin: ", as.character(gbegin), "; bbegin: ", as.character(bbegin), "; abegin: ", as.character(abegin), ".\n"))
		#cat(paste(sep="", "rend: ", as.character(rend), "; gend: ", as.character(gend), "; bend: ", as.character(bend), "; aend: ", as.character(aend), ".\n"))

		## Create the intermediate color
		rout = rbegin * (1 - r) + rend * r
		gout = gbegin * (1 - r) + gend * r
		bout = bbegin * (1 - r) + bend * r
		aout = abegin * (1 - r) + aend * r

		#cat(paste(sep="", "rout: ", as.character(rout), "; gout: ", as.character(gout), "; bout: ", as.character(bout), "; aout: ", as.character(aout), ".\n"))

		rr = ratio(x, yy, i)
		gbegin = segment(x, yy, i, TRUE)
		gend = segment(x, yy, i, FALSE)
		#cat(paste(sep = "", "i: ", as.character(i), "; x: ", as.character(x), "; yy = ", as.character(yy), "; gbegin: ", as.character(gbegin), "; gend: ", as.character(gend), "; rr: ", as.character(rr)))

		## Integrate luminosity into the gradient, for grayscale outputs
		hsvbegin = rgb2hsv(r = col2rgb(grayscale[gbegin], alpha = TRUE)["red",], g = col2rgb(grayscale[gbegin], alpha = TRUE)["green",], b = col2rgb(grayscale[gbegin], alpha = TRUE)["blue",], maxColorValue = 255)
		hsvend = rgb2hsv(r = col2rgb(grayscale[gend], alpha = TRUE)["red",], g = col2rgb(grayscale[gend], alpha = TRUE)["green",], b = col2rgb(grayscale[gend], alpha = TRUE)["blue",], maxColorValue = 255)
		hsvout = rgb2hsv(r = rout, g = gout, b = bout, maxColorValue = 255)

		hsvout[3] = hsvbegin[3] * (1 - rr) + hsvend[3] * rr

		out = append(out, hsv(hsvout[1], hsvout[2], hsvout[3], aout / 255))
	}

	return(unlist(out))
}

## Largely inspired from R built-in summarize function. This one duplicates results as much as there are rows in each group
spread = function(.data, ...) {
	stopifnot(is.data.frame(.data) || is.list(.data) || is.environment(.data))
	cols <- as.list(substitute(list(...))[-1])

	if (is.null(names(cols))) {
		missing_names <- rep(TRUE, length(cols))
	}
	else {
		missing_names <- names(cols) == ""
	}

	if (any(missing_names)) {
		names <- unname(unlist(lapply(match.call(expand.dots = FALSE)$..., deparse)))
		names(cols)[missing_names] <- names[missing_names]
	}

	for (col in names(cols)) {
		.data[,col] <- rep(eval(cols[[col]], .data, parent.frame()), length(.data[,names(.data)[1]]))
	}

	return(.data)
}

plotit = function(input, num, name, x, y, labels, xaxis, yaxis, summary, filter, normalize)
{
	cat(paste("Plotting graphs ", as.character(num), ", ", as.character(num + 1), " and ", as.character(num + 2), ": ", name, " by ", x, ".\n", sep=""))

	if(length(input[,x]) <= 0)
	{
		cat("No data available for this plot. Skipping.\n")
		return(num + 3)
	}

	## Use this to debug the plotting process of a particular figure using its number. Disable or enable it with "&& 0" or "&& 1", respectively.
	if(num != 1 && 0)
	{
		return(num + 3);
	}

	if(filter)
	{
		## Filter out all lines where at least one quantity represented by a curve exhibits a unfeasible solution
		## Or schedulers or evaluators failed to produce a correct output
		## Unfeasible solutions are considered energy = 0, which brings some unfair advantage to these schedulers
		## * Groups lines in clusters of x and other-than-x variables combinations
		## * Compute the mean value for field "feasible" within each cluster (same for scheduler and evaluation exit values)
		## * Filter out all ines whose new field "feasible" is strictly lower than 1 (and lines whose average schedule exit value or evaluation exit value is strictly higher than 0)
		## Apply the same technique for schedule and evaluation exit values
		columns = levels(as.factor(c(x, "application", "p", "evaluation", "evaluation_try", "schedule_try")))
		
		input = ddply(input, columns, spread, feasible = sum(feasible) / length(feasible), valid_schedule = mean(valid_schedule))
		input = input[input$feasible == 1 & input$valid_schedule == 1,]
	}

	## Compute mean and standard deviation in the remaining data
	input = ddply(input, c("scheduler", x), here(summarize), mean = mean(eval(parse(text = y))), sd = sd(eval(parse(text = y))), feasible = mean(feasible), valid_schedule = mean(valid_schedule))
	input$sd[is.na(input$sd)] = 0

	if(normalize)
	{
		minval = min(input$mean)
		ylabel_notice = "( "
		if(minval < 1)
		{
			cat("Warning: Applying logarithmic reduction to values lower than 1. Multiplying values by", minval, "first.\n")
			input$mean = input$mean * (1 / minval)
			ylabel_notice = paste(ylabel_notice, "x", minval, ", ")
		}
		ylabel_notice = paste(ylabel_notice, "log scale)")
		input = ddply(input, c("scheduler", x), here(spread), mean = log(mean), sd = mean(sd), feasible = mean(feasible), valid_schedule = mean(valid_schedule))
	}
	else
	{
		ylabel_notice = ""
	}

	## Now generate three plots: error bar first, then simple curves and finally a bar graph
	## Write raw output
	write.csv(input, paste(sep = "", name, ".csv"))

	## Create a base plot object
	plotnr = num + 1;
	plot_base = ggplot() + theme[[plotnr]] + 
		guides(color = guide_legend(title = "Scheduler"), shape = guide_legend(title = "Scheduler"), fill = guide_legend(title = "Scheduler")) +
		ylab(paste(yaxis, ylabel_notice)) +
		xlab(xaxis) +
		ggtitle(summary) + 
		scale_color_manual(values = gradient(number = length(unique(input$scheduler)) + 1)) +
		scale_fill_manual(values = gradient(number = length(unique(input$scheduler)) + 1))

	## Add lines and points
	plot = plot_base + 
		geom_line(data = apply_labels(input),
			aes_string(ifelse(group[plotnr], paste(sep = "", "as.factor(", x, ")"), paste(sep = "", "type.convert(as.character(", x, "), as.is = TRUE)")), "mean", group = "scheduler", color = "scheduler", shape = "scheduler"),
			size = thickness[plotnr]
		) +
		geom_point(data = apply_labels(input),
			aes_string(ifelse(group[plotnr], paste(sep = "", "as.factor(", x, ")"), paste(sep = "", "type.convert(as.character(", x, "), as.is = TRUE)")), "mean", group = "scheduler", color = "scheduler", shape = "scheduler"),
			size = marker_size[plotnr]
		)
	## Save the plot as a svg filie
	ggsave(file = paste(sep = "", "plot-", as.character(plotnr), "-", output_prefix[plotnr], "-", name, ".", output_extension[plotnr]), plot = plot, width = 8, height = 6)


	## Add error bars to the previous plot
	plotnr = num;
	error = plot + 
		geom_errorbar(data = apply_labels(input),
			aes_string(ifelse(group[plotnr], paste(sep = "", "as.factor(", x, ")"), paste(sep = "", "type.convert(as.character(", x, "), as.is = TRUE)")), "mean", ymin = "mean - sd", ymax = "mean + sd", group = "scheduler", color = "scheduler", shape = "scheduler"),
			size = thickness[plotnr]
		)
	## Save as a svg file
	ggsave(file = paste(sep = "", "error-", as.character(plotnr), "-", output_prefix[plotnr], "-", name, ".", output_extension[plotnr]), plot = error, width = 8, height = 6)

	## Now create a bar chart
	plotnr = num + 2;
	bars = plot_base +
		geom_bar(data = apply_labels(input),
			aes_string(ifelse(!group[plotnr], paste(sep = "", "as.factor(", x, ")"), paste(sep = "", "type.convert(as.character(", x, "), as.is = TRUE)")), "mean", group = "scheduler", fill = "scheduler", shape = "scheduler"),
			position = "dodge", stat = "identity")
	## Save as a svg file
	ggsave(file = paste(sep = "", "bar-", as.character(plotnr), "-", output_prefix[plotnr], "-", name, ".", output_extension[plotnr]), plot = bars, width = 8, height = 6)

	return(num + 3)
}

source("settings.r")

#x11()
#m = matrix(1:10 / 10, 10, 1)
#image(m, col = gradient())
##invisible(readLines("stdin", n=1))
#stopifnot(FALSE)

## First compute the energy from the power function (just integrate it on [0,max(time)]
data.frame = ddply(
	data.frame, c("scheduler", "application", "p", "evaluation", "schedule_try", "evaluation_try", "n", "feasible", "n"), summarize,
	energy = as.numeric(integrate(approxfun(time, power, method = "linear"), subdivisions = 10000L, min(time), max(time))[1]),
	time_global = mean(time_global),
	time_mapping = mean(time_mapping),
	time_scaling = mean(time_scaling),
	work_stddev = mean(work_stddev),
	max_width_mean = mean(max_width_mean),
	max_width_stddev = mean(max_width_stddev),
	valid_schedule = mean(valid_schedule)
)

## Keep only the first evaluation method, as the block below is evaluation-independent 
schedule = data.frame[data.frame[,"evaluation"] == data.frame$evaluation[1],]

num = 1
num = plotit(schedule, num, "time_global_tasks", "n", "time_global", c(), "Number of tasks", "Time in seconds", "Time to compute a schedule", TRUE, FALSE)
num = plotit(schedule, num, "time_global_cores", "p", "time_global", c(), "Number of cores", "Time in seconds", "Time to compute a schedule", TRUE, FALSE)
num = plotit(schedule, num, "time-global_class", "application", "time_global", c(), "Task class", "Time in seconds", "Time to compute a schedule", TRUE, FALSE);

for(i in unique(data.frame$evaluation))
{
	schedule = data.frame[data.frame$evaluation == i,];

	num = plotit(schedule, num, paste(sep = "", "feasible_tasks", "-", i), "n", "feasible", c(), "Number of tasks", "success rate", "Schedulers' success rate", FALSE, FALSE);
	num = plotit(schedule, num, paste(sep = "", "feasible_cores", "-", i), "p", "feasible", c(), "Number of cores", "success rate", "Schedulers' success rate", FALSE, FALSE);
	num = plotit(schedule, num, paste(sep = "", "feasible_class", "-", i), "application", "feasible", c(), "Task class", "success rate", "Schedulers' success rate", FALSE, FALSE);

	num = plotit(schedule, num, paste(sep = "", "valid_tasks", "-", i), "n", "valid_schedule", c(), "Number of tasks", "success rate", "Schedulers' success rate", FALSE, FALSE);
	num = plotit(schedule, num, paste(sep = "", "valid_cores", "-", i), "p", "valid_schedule", c(), "Number of cores", "success rate", "Schedulers' success rate", FALSE, FALSE);
	num = plotit(schedule, num, paste(sep = "", "valid_class", "-", i), "application", "valid_schedule", c(), "Task class", "success rate", "Schedulers' success rate", FALSE, FALSE);

	num = plotit(schedule, num, paste(sep = "", "energy_tasks", "-", i), "n", "energy", c(), "Number of tasks", "Energy", paste(sep = "", "Energy quality of schedules"), TRUE, normalize_energy);
	num = plotit(schedule, num, paste(sep = "", "energy_cores", "-", i), "p", "energy", c(), "Number of cores", "Energy", paste(sep = "", "Energy quality of schedules"), TRUE, normalize_energy);
	num = plotit(schedule, num, paste(sep = "", "energy_class", "-", i), "application", "energy", c(), "Task class", "Energy", paste(sep = "", "Energy quality of schedules"), TRUE, normalize_energy);
	max_stddev = schedule[schedule$work_stddev > 0,];
	num = plotit(max_stddev, num, paste(sep = "", "energy_max_width", "-", i), "max_width_mean", "energy", c(), "Maximum task width mean in task set", "Energy", paste(sep = "", "Energy quality"), TRUE, normalize_energy);

	frequency = schedule[schedule$scheduler == "crown_ilp_balanced_ilp_simple" || schedule$scheduler == "crown_ilp_balanced_ilp_idle" || schedule$scheduler == "crown_ilp_balanced_ilp_off" || schedule$scheduler == "crown_ilp_balanced_height",];

	num = plotit(frequency, num, paste(sep = "", "energy-frequency_tasks", "-", i), "n", "energy", c(), "Number of tasks", "Energy", paste(sep = "", "Schedule energy"), TRUE, normalize_energy);
	num = plotit(frequency, num, paste(sep = "", "energy-frequency_cores", "-", i), "p", "energy", c(), "Number of cores", "Energy", paste(sep = "", "Schedule energy"), TRUE, normalize_energy);
	num = plotit(frequency, num, paste(sep = "", "energy-frequency_class", "-", i), "application", "energy", c(), "Task class", "Energy", paste(sep = "", "Schedule energy"), TRUE, normalize_energy);
	max_stddev = frequency;
	num = plotit(max_stddev, num, paste(sep = "", "energy-frequency_max_width", "-", i), "max_width_mean", "energy", c(), "Maximum task width mean in task set", "Energy", paste(sep = "", "Schedule energy"), TRUE, normalize_energy);
}
